#!/usr/bin/python3
import re
import os
import pathos
import subprocess
from collections import namedtuple
from matplotlib import pyplot

def run_evo(commands):
  p = subprocess.Popen(commands, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
  p.wait()
  stdout, stderr = p.communicate()

  if len(stderr):
    print(stderr.decode('utf-8'))

  result = stdout.decode('utf-8')
  results = {}
  for item in re.findall(r'([a-z]+)\s+([0-9]+\.[0-9]+)', result):
    results[item[0]] = float(item[1])

  return results


def eval_ape(gt_filename, traj_filename, t_offset=0.0):
  ret = run_evo(['evo_ape', 'kitti', gt_filename, traj_filename, '-a'])
  return ret


def eval_rpe(gt_filename, traj_filename, delta_unit='m', delta=100, all_pairs=True, t_offset=0.0):
  commands = ['evo_rpe', 'kitti', gt_filename, traj_filename, '-a', '--delta_unit', str(delta_unit), '--delta', str(delta)]
  if all_pairs:
    commands += ['--all_pairs']

  ret = run_evo(commands)
  return ret


def main():
  gt_path = '/home/koide/datasets/ssd/kitti/poses/00_lidar.txt'

  results_path = os.path.dirname(__file__) + '/results'
  
  filenames = []
  for filename in os.listdir(results_path):
    found = re.findall(r'traj_lidar_(\S+)_(\d+).txt', filename)
    if not found:
      continue
    
    method = found[0][0] + '_' + found[0][1]
    filenames.append((method, results_path + '/' + filename))
  
  methods = ['pcl_1', 'fast_gicp_128', 'fast_vgicp_128', 'small_gicp_1', 'small_gicp_tbb_128', 'small_gicp_omp_128', 'small_vgicp_tbb_128']
  labels = ['pcl_gicp', 'fast_gicp', 'fast_vgicp', 'small_gicp', 'small_gicp (tbb)', 'small_gicp (omp)', 'small_vgicp']
  
  Result = namedtuple('Result', ['ape', 'rpe100', 'rpe400', 'rpe800'])
  def evaluate(inputs):
    method, filename = inputs
    print('.', end='', flush=True)
    ape = eval_ape(gt_path, filename)
    rpe100 = eval_rpe(gt_path, filename, delta=100)
    rpe400 = eval_rpe(gt_path, filename, delta=400)
    rpe800 = eval_rpe(gt_path, filename, delta=800)
    return method, Result(ape, rpe100, rpe400, rpe800)

  print('evaluating')
  with pathos.multiprocessing.ProcessingPool() as p:
    errors = p.map(evaluate, [(method, filename) for method, filename in filenames if method in methods])
  print()
  
  results = {}
  for method, error in errors:
    results[method] = error
    
  for method, label in zip(methods, labels):
    ape, rpe100, rpe400, rpe800 = results[method]
    print('{:20s} : APE={:.3f} +- {:.3f}  RPE(100)={:.3f} +- {:.3f}  RPE(400)={:.3f} +- {:.3f}  RPE(800)={:.3f} +- {:.3f}'
          .format(label, ape['rmse'], ape['std'], rpe100['rmse'], rpe100['std'], rpe400['rmse'], rpe400['std'], rpe800['rmse'], rpe800['std']))
  
  fig, axes = pyplot.subplots(1, 3, figsize=(15, 5))
  apes = [results[method].ape['rmse'] for method in methods]
  axes[0].bar(labels, apes)
  
  pyplot.show()
  
    
  
  

if __name__ == '__main__':
  main()



#!/usr/bin/python3
import re
import os
import numpy
from matplotlib import pyplot
from collections import namedtuple

def parse_result(filename):
  num_points = []
  results = {}
  
  with open(filename, 'r') as f:
    for line in f.readlines():
      matched = re.match(r'num_threads=(\d+)', line)
      if matched:
        num_threads = int(matched.group(1))
        continue
      
      matched = re.match(r'num_points=(\d+)', line)
      if matched:
        num_points.append(int(matched.group(1)))
        continue
      
      matched = re.match(r'(\S+)_times=(\S+) \+\- (\S+) \(median=(\S+)\)', line)
      if matched:
        method = matched.group(1)
        mean = float(matched.group(2))
        std = float(matched.group(3))
        med = float(matched.group(4))

        if method not in results:
          results[method] = []
        
        results[method].append(med)
        continue

  return (num_threads, num_points, results)

def main():
  results_path = os.path.dirname(__file__) + '/results'

  results = {}
  for filename in os.listdir(results_path):
    matched = re.match(r'kdtree_benchmark_(\S+)_(\d+).txt', filename)
    if not matched:
      continue

    method = '{}_{}'.format(matched.group(1), matched.group(2))
    print(method)

    num_threads, num_points, rets = parse_result(results_path + '/' + filename)
    results[method] = rets[list(rets.keys())[0]] 

  fig, axes = pyplot.subplots(1, 2, figsize=(12, 3))
  
  num_threads = [1, 2, 3, 4, 5, 6, 7, 8, 16, 32, 64, 128]
  axes[0].plot(num_points, results['small_1'], label='kdtree (single-thread)', marker='o', linestyle='--')
  for idx in [1, 2, 3, 5, 7, 8, 9]:
    N = num_threads[idx]
    axes[0].plot(num_points, results['omp_{}'.format(N)], label='kdtree_omp (%d threads)' % N, marker='s')
    # axes[0].plot(num_points, results['tbb_{}'.format(N)], label='kdtree_tbb (%d threads)' % N, marker='^')

  baseline = numpy.array(results['small_1'])
  axes[1].plot([num_threads[0], num_threads[-1]], [1.0, 1.0], label='kdtree (single-thread)', linestyle='--')
  for idx in [5]:
    threads = num_threads[idx]
    N = num_points[idx]
    
    axes[1].plot(num_threads, baseline[idx] / [results['omp_{}'.format(threads)][idx] for threads in num_threads], label='omp (num_points=%d)' % N, marker='s')
    axes[1].plot(num_threads, baseline[idx] / [results['tbb_{}'.format(threads)][idx] for threads in num_threads], label='tbb (num_points=%d)' % N, marker='^')

  axes[1].set_xscale('log')
   
  axes[0].set_xlabel('Number of points')
  axes[0].set_ylabel('KdTree construction time [msec/scan]')
  axes[1].set_xlabel('Number of threads = [1, 2, 4, ..., 128]')
  axes[1].set_ylabel('Processing speed ratio (single-thread = 1.0)')

  axes[0].grid()
  axes[1].grid()
  
  axes[0].legend()
  axes[1].legend()
  
  fig.savefig('kdtree_time.svg')
  
  pyplot.show()
  
  

if __name__ == '__main__':
  main()


#!/usr/bin/python3
import os
import time
import argparse
import collections
import numpy
import small_gicp
from pyridescence import *

# Odometry estimation based on scan-to-scan matching
class ScanToScanMatchingOdometry(object):
  def __init__(self, num_threads):
    self.num_threads = num_threads
    self.T_last_current = numpy.identity(4)
    self.T_world_lidar = numpy.identity(4)
    self.target = None
  
  def estimate(self, raw_points):
    downsampled, tree = small_gicp.preprocess_points(raw_points, 0.25, num_threads=self.num_threads)
    
    if self.target is None:
      self.target = (downsampled, tree)
      return self.T_world_lidar

    result = small_gicp.align(self.target[0], downsampled, self.target[1], self.T_last_current, num_threads=self.num_threads)
    
    self.T_last_current = result.T_target_source
    self.T_world_lidar = self.T_world_lidar @ result.T_target_source
    self.target = (downsampled, tree)
    
    return self.T_world_lidar
    
# Odometry estimation based on scan-to-model matching
class ScanToModelMatchingOdometry(object):
  def __init__(self, num_threads):
    self.num_threads = num_threads
    self.T_last_current = numpy.identity(4)
    self.T_world_lidar = numpy.identity(4)
    self.target = small_gicp.GaussianVoxelMap(1.0)
    self.target.set_lru(horizon=100, clear_cycle=10)
  
  def estimate(self, raw_points):
    downsampled, tree = small_gicp.preprocess_points(raw_points, 0.25, num_threads=self.num_threads)
    
    if self.target.size() == 0:
      self.target.insert(downsampled)
      return self.T_world_lidar
    
    result = small_gicp.align(self.target, downsampled, self.T_world_lidar @ self.T_last_current, num_threads=self.num_threads)

    self.T_last_current = numpy.linalg.inv(self.T_world_lidar) @ result.T_target_source
    self.T_world_lidar = result.T_target_source
    self.target.insert(downsampled, self.T_world_lidar)
    
    guik.viewer().update_drawable('target', glk.create_pointcloud_buffer(self.target.voxel_points()[:, :3]), guik.Rainbow())
    
    return self.T_world_lidar
  

def main():
  parser = argparse.ArgumentParser()
  parser.add_argument('dataset_path', help='/path/to/kitti/velodyne')
  parser.add_argument('--num_threads', help='Number of threads', type=int, default=4)
  parser.add_argument('-m', '--model', help='Use scan-to-model matching odometry', action='store_true')
  args = parser.parse_args()
  
  dataset_path = args.dataset_path
  filenames = sorted([dataset_path + '/' + x for x in os.listdir(dataset_path) if x.endswith('.bin')])
  
  if not args.model:
    odom = ScanToScanMatchingOdometry(args.num_threads)
  else:
    odom = ScanToModelMatchingOdometry(args.num_threads)
  
  viewer = guik.viewer()
  viewer.disable_vsync()
  time_queue = collections.deque(maxlen=500)

  for i, filename in enumerate(filenames):
    raw_points = numpy.fromfile(filename, dtype=numpy.float32).reshape(-1, 4)[:, :3]
    
    t1 = time.time()
    T = odom.estimate(raw_points)
    t2 = time.time()
   
    time_queue.append(t2 - t1)
    viewer.lookat(T[:3, 3])
    viewer.update_drawable('points', glk.create_pointcloud_buffer(raw_points), guik.FlatOrange(T).add('point_scale', 2.0))
    
    if i % 10 == 0:
      viewer.update_drawable('pos_{}'.format(i), glk.primitives.coordinate_system(), guik.VertexColor(T))
      viewer.append_text('avg={:.3f} msec/scan  last={:.3f} msec/scan'.format(1000 * numpy.mean(time_queue), 1000 * time_queue[-1]))
    
    if not viewer.spin_once():
      break    

if __name__ == '__main__':
  main()
